function varargout = hmex (varargin)
% Demonstrate how to use hmmvp with practically useful Green's functions.
%   I have not used these Green's function implementations in simulations, so
% they are not nearly as well tested as, say, dc3dm. If you use these as a model
% for your own code, let me know if you find any bugs.
%   From the command line, run examples like this:
%      >> c = hmex('ok_ex1_build');
%      >> hmex('ok_ex1_mvp', c);
%   To see a description of each Green's function, in a shell type
%      $ ./bin/hmmvpbuild help compress
  [varargout{1:nargout}] = feval(varargin{:});
end

function gp = get_global_parms ()
% Modify these for your system.
  % hmmvp base directory.
  gp.hmmvphome = '/scratch/ambrad/release/hmmvp_v1.3';
  % Output directory.
  gp.odir = '/scratch/ambrad/hmex';
  % Location of hmmvpbuild. This is probably fine as is.
  fp = mfilename('fullpath');
  [p n e] = fileparts(fp);
  gp.hmmvpbuild = [p filesep '../bin/hmmvpbuild'];
  % Location of distmesh.
  gp.distmesh = [p filesep '../extern/distmesh'];
  % Number of OpenMP threads to use.
  gp.nthreads = 4;
end

function addpaths ()
  gp = get_global_parms();
  addpath([gp.hmmvphome filesep 'matlab']);
  addpath(gp.distmesh);
end

% ------------------------------------------------------------------------------
% For use with greens_fn 'okada'.
%   This example illustrates one way to encode a problem in which a few planar
% faults are discretized.

% -- examples
function c = ok_ex1_build ()
% Example to run from the command line: c = hmex('ok_ex1_build')
  o = ok_setopts();
  o.rfac = 5;
  o.vecs = o.vecs(:, [2 3]);

  rfacs = [1 o.rfac];
  for (i = 1:numel(rfacs))
    % Make just one rectangular fault.
    oms = {ok_mesh_rectangle(4, 2, rfacs(i)*8, rfacs(i)*4, [-2 1 -2], 20, 12)};
    if (i == 1) ok_drawmesh(oms); drawnow; end
  end
  
  c = ok_build_hmatrices(oms, o);
end
function ok_ex1_mvp (c, k)
% Example.
  if (nargin < 2) k = 1; end
  id = hmmvp('init', c.hms{k});
  om = c.oms{1};
  slip = tapered_slip_fn(om.Xc, om.Yc, 0.8, 3);
  tau = hmmvp('mvp', id, slip(:));
  hmmvp('cleanup', id);
  tau = reshape(tau, size(slip));
  subplot(211); imagesc(om.xc, om.yc, slip); colorbar; axis equal; axis tight;
  subplot(212); imagesc(om.xc, om.yc, tau); colorbar; axis equal; axis tight;
end

function c = ok_ex2 ()
% Example.
  o = ok_setopts();
  o.vecs = o.vecs(:,[1]);
  o.rfac = 5;
  % Build one matrix for each (source fault, receiver fault, source component,
  % receiver component) 4-tuple.
  oms = {ok_mesh_rectangle(4, 2, o.rfac*8, o.rfac*4, [-2 1 -15], 20, 60), ...
         ok_mesh_rectangle(3, 2, o.rfac*6, o.rfac*4, [1 -1 -13], -10, 80)};
  ok_drawmesh(oms);
  c = ok_build_hmatrices(oms, o);
end

% -- public
function o = ok_setopts ()
  o.tol = 1e-5;
  % Resolution factor. Increase to make a finer mesh.
  o.rfac = 1;
  % Vectors in element-local coordinates that describe dislocation and receiver
  % traction vectors. The usual case is eye(3), which is along-strike,
  % along-dip, and fault-normal.
  o.vecs = eye(3);
  % Fullspace rather than halfspace.
  o.want_fullspace = 0;
end

function om = ok_mesh_rectangle (Dx, Dy, nx, ny, ctr, striked, dipd)
% Mesh a rectangle nx by ny mesh having along-strike length Dx and along-dip
% length Dy. Place it in global space according to the center, strike, and dip.
  nr = nx*ny;
  xc = CC(linspace(-0.5*Dx, 0.5*Dx, nx+1));
  yc = CC(linspace(-0.5*Dy, 0.5*Dy, ny+1));
  [Xc Yc] = meshgrid(xc, yc);
  om.ctr = ok_transform(ctr, dipd, striked, [Xc(:).'; Yc(:).'; zeros(1, nr)]);
  one = ones(1, nr);
  om.striked = striked*one;
  om.dipd = dipd*one;
  om.las = Dx/nx*one;
  om.lad = Dy/ny*one;
  % For convenience:
  om.xc = xc; om.yc = yc;
  om.Xc = Xc; om.Yc = Yc;
end

function h = ok_drawmesh (om)
% Draw the mesh using lines in 3D.
  if (iscell(om))
    clr = 'bgrkcm';
    for (i = 1:numel(om))
      h = ok_drawmesh(om{i});
      set(h, 'color', clr(mod(i-1, numel(clr)) + 1));
    end
    return;
  end
  nr = size(om.ctr, 2);
  ls = cell(1, 3); [ls{1} ls{2} ls{3}] = deal(zeros(2, 4*nr));
  for (i = 1:nr)
    vs = ok_calc_rect_vertices( ...
      om.ctr(:,i), om.dipd(i), om.striked(i), om.las(i), om.lad(i));
    cs = 4*(i - 1) + (1:4);
    for (j = 1:3) ls{j}(:, cs) = [vs(j, 1:4); vs(j, [2:4, 1])]; end
  end
  h = line(ls{1}, ls{2}, ls{3}, 'color', 'b');
end

function c = ok_build_hmatrices (oms, o)
% Build H-matrices for 3-component interactions among all faults.
  if (~iscell(oms)) oms = {oms}; end
  % Clean up memory in case an H-matrix is hanging around.
  clear hmmvp;
  
  c.o = o;
  c.oms = oms;

  % Each traction component is the sum of multiple matrix-vector products:
  %     tau_i = sum_j B_ij x_ij.
  % Control the error on tau_i using max_j ||B_ij||_F.
  [c.hms c.comps c.cf] = ok_compress(c);
end

% -- private
function Bfro = ok_read_Bfroest (fn)
  fid = fopen(sprintf('%s.Bfroest', fn), 'r');
  if (fid == -1) error(sprintf('Could not open %s.Bfroest', fn)); end
  Bfro = fscanf(fid, '%e');
  fclose(fid);
end

function [hms comps cf] = ok_compress (c, Bfro)
% 'if' is the fault index into c.oms. 'ic' is the component index, in this case
% into along-slip, along-dip, and fault-normal.
%   hms is a 4D cell array indexed as follows:
%     hms{source fault, receiver fault,
%         source dislocation index, receiver vector index}.
%   comps(i,:) contains this 4-tuple for hms{i}.

  % Compress the self-interaction matrices.
  Bfro = zeros(numel(c.oms), numel(c.oms), ...
               size(c.o.vecs, 2), size(c.o.vecs, 2));
  cf = Bfro;
  for (io = 1:numel(c.oms))
    for (iv = 1 : size(c.o.vecs, 2))
      comps = {io, io, iv, iv};
      % Compress this self-interaction matrix.
      tmp_fn = make_tmp_filename();
      k = ok_make_original_kvf(c, [], tmp_fn, comps{:});
      run_hmmvp_build(k);
      % Determine the ||B_ij||_F essentially exactly.
      id = hmmvp('init', k.write_hmat_filename);
      % All matrices B_ij contributing to this traction receiver tau_i use the
      % same ||B_ij||_F.
      Bfro(:, io, :, iv) = sqrt(hmmvp('fronorm2', id));
      hmmvp('cleanup', id);
      % Recompress.
      fn = ok_make_filename(c.o, comps{:});
      k = ok_make_recompress_kvf(k, Bfro(comps{:}), fn);
      hms{comps{:}} = k.write_hmat_filename;
      cf(io, io, iv, iv) = run_hmmvp_build(k);
      % Remove the original H-matrix.
      rm_tmp_files(tmp_fn);
    end
  end
  
  % Compress the self-other-interaction matrices.
  for (ios = 1:numel(c.oms))
    for (ior = 1:numel(c.oms))
      for (ivs = 1 : size(c.o.vecs, 2))
        for (ivr = 1 : size(c.o.vecs, 2))
          if (ios == ior && ivs == ivr) continue; end
          comps = {ios, ior, ivs, ivr};
          fn = ok_make_filename(c.o, comps{:});
          k = ok_make_original_kvf(c, Bfro(comps{:}), fn, comps{:});
          k = rmfield(k, 'write_hd_filename');
          hms{comps{:}} = k.write_hmat_filename;
          cf(comps{:}) = run_hmmvp_build(k);
        end
      end
    end
  end
  
  comps = zeros(numel(hms), 4);
  [comps(:,1), comps(:,2), comps(:,3), comps(:,4)] = ...
      ind2sub(size(hms), 1:numel(hms));
end

function fn = ok_make_filename (o, ifsrc, ifrcv, icsrc, icrcv)
  gp = get_global_parms();
  fn = sprintf(rmsp('%s/ok_ rfac%1.1f tol%d sf%d rf%d sc%d rc%d'), ...
               gp.odir, o.rfac, -round(log10(o.tol)), ifsrc, ifrcv, icsrc, ...
               icrcv);
end

function k = ok_make_original_kvf (c, Bfro, fn, ifsrc, ifrcv, icsrc, icrcv)
% Write the key-value file for hmmvpbuild.
  gp = get_global_parms();
  
  % For hmmvpbuild.
  k.write_hmat_filename = fn;
  k.write_hd_filename = fn;
  k.greens_fn = 'okada';
  k.tol = c.o.tol;
  k.allow_overwrite = 1;
  k.nthreads = gp.nthreads;
  k.want_fullspace = c.o.want_fullspace;
  if (~isempty(Bfro)) k.Bfro = Bfro; end
  
  % For 'okada' Green's function.
  if (ifsrc == ifrcv)
    om = c.oms{ifsrc};
    k.src_idxs = 1:numel(om.dipd);
    k.rcv_idxs = k.src_idxs;
  else
    om = ok_merge_meshes(c.oms([ifsrc ifrcv]));
    k.src_idxs = 1:numel(c.oms{ifsrc}.dipd);
    k.rcv_idxs = k.src_idxs(end) + (1:numel(c.oms{ifrcv}.dipd));
  end
  assert(numel(unique([k.src_idxs, k.rcv_idxs])) == numel(om.dipd));

  k.mu = 1;
  k.nu = 0.25;
  k.center = om.ctr;
  k.striked = om.striked;
  k.dipd = om.dipd;
  k.len_along_dip = om.lad;
  k.len_along_strike = om.las;
  k.disl = repmat(c.o.vecs(:,icsrc), 1, numel(om.dipd));
  k.rcv  = repmat(c.o.vecs(:,icrcv), 1, numel(om.dipd));
end

function k = ok_make_recompress_kvf (k, Bfro, fn)
% Modify the key-value file to recompress to the final H-matrix file.
  k.use_hmat_filename = k.write_hmat_filename;
  k.write_hmat_filename = fn;
  k.use_hd_filename = k.write_hd_filename;
  k = rmfield(k, 'write_hd_filename');
  k.Bfro = Bfro;
end

function v = ok_transform (ctr, dipd, striked, v)
  c = cosd(dipd); s = sind(dipd);
  v = [1 0 0; 0 c -s; 0 s c] * v;
  c = cosd(striked); s = sind(striked);
  v = [c -s 0; s c 0; 0 0 1]' * v;
  v = repmat(ctr(:), 1, size(v, 2)) + v;
end
function v = ok_calc_rect_vertices (ctr, dipd, striked, las, lad)
% Angle conventions:
%   striked is relative to N. +ve makes it rotate CW.
%   dipd is relative to horizontal. +ve makes the N side move up.
  v = 0.5*[-1 -1 0; 1 -1 0; 1 1 0; -1 1 0]';
  v = repmat([las; lad; 0], 1, 4) .* v;
  v = ok_transform(ctr, dipd, striked, v);
end

function om = ok_merge_meshes (oms)
% Make one om struct from a cell array of multiple ones.
  om = oms{1};
  flds = {'ctr', 'striked', 'dipd', 'las', 'lad'};
  for (i = 2:numel(oms))
    for (j = 1:numel(flds))
      om.(flds{j}) = [om.(flds{j}), oms{i}.(flds{j})];
    end
  end
end

% ------------------------------------------------------------------------------
% For use with greens_fn 'tgf'.

% -- examples
function c = tm_ex1_build ()
% Example.
  o = tm_setopts();
  o.rfac = 3;
  
  xlim = [-1 1]; ylim = [-1 1];
  dx = diff(xlim)/(6*o.rfac); dy = diff(ylim)/(6*o.rfac);
  dipd = 12; min_depth = 2;
  types = [2 4];
  t = tm_mesh_rectangle(xlim, ylim, dx, dy, dipd, min_depth);
  tm_drawmesh(t); axis equal; axis tight;

  c = tm_build_hmatrices(t, o);
end
function tm_ex1_mvp (c)
% Example.
  x = linspace(-1, 1, 100);
  
  slip = tapered_slip_fn(c.t.centroid(1,:), c.t.centroid(2,:), 0.8, 3);
  h(1) = sp(221); tm_drawmesh(c.t, slip); view(2); cb;
  h(2) = sp(222); imagesc(x, x, tsi_interp(tsi_make(c.t, slip), x, x)); cb;
  axis xy;
  
  id = hmmvp('init', c.hm);
  tau = hmmvp('mvp', id, slip(:));
  hmmvp('cleanup', id);
  h(3) = sp(223); tm_drawmesh(c.t, tau); view(2); cb;
  h(4) = sp(224); imagesc(x, x, tsi_interp(tsi_make(c.t, tau), x, x)); cb;
  axis xy;
  
  linkaxes(h, 'xy');
end

% -- public
function o = tm_setopts ()
  o.tol = 1e-5;
  % Resolution factor. Increase to make a finer mesh.
  o.rfac = 1;
  % Fullspace rather than halfspace.
  o.want_fullspace = 0;
  % Dislocation in local coordinates. x is along strike, y is along dip, and z is
  % element normal.
  o.disl = [0 1 0];
  % Receiver traction vector in local coordinates.
  o.rcv = o.disl;
end

function t = tm_mesh_rectangle (xlim, ylim, dx, dy, dipdeg, min_depth)
  % Use distmesh2d to mesh the fault.
  h0 = 0.5*(dx + dy);
  xlim = xlim(:); ylim = ylim(:);
  bbox = [xlim, ylim];
  pfix = [xlim([1 2 2 1]), ylim([1 1 2 2])];
  fd = @(p) drectangle(p, xlim(1), xlim(2), ylim(1), ylim(2));
  fh = @huniform;
  [t.v t.tri] = distmesh2d(fd, fh, h0, bbox, pfix);
  t.v = t.v';
  t.v(3,:) = 0;
  t.tri = t.tri';
  t = tm_transform(t, dipdeg, min_depth);
  t = tm_calc_centroid(t);
end

function tm_drawmesh (t, q)
  if (nargin < 2 || isempty(q)) q = 1:size(t.tri, 2); end
  if (0)
    h = trimesh(t.tri', t.v(1,:), t.v(2,:), t.v(3,:), q);
    set(h, 'edgecolor', 'r');
  else
    trisurf(t.tri', t.v(1,:), t.v(2,:), t.v(3,:), q);
  end
end

function c = tm_build_hmatrices (t, o)
  % Clean up memory in case an H-matrix is hanging around.
  clear hmmvp;

  c.o = o;
  c.t = t;
  
  [c.hm c.cf] = tm_compress(c);
end

% ------------------------------------------------------------------------------
% For use with greens_fn 'btrmbem'.

function c = btrm_ex1 ()
  c = btrm_make_problem();

  subplot(211);
  dxb = diff(c.xb);
  plot(CC(c.xb), log2(dxb/min(dxb)), '.-');
  xlabel('x'); ylabel('log_2 \Deltax');
  
  [c.Gs c.Gn] = btrm_call_hmmvp (c);
  
  subplot(223); imagesc(log10(abs(c.Gs))); colorbar; title('G_s');
  subplot(224); imagesc(log10(abs(c.Gn))); colorbar; title('G_n');
end

function c = btrm_make_problem ()
  c.xb = btrm_make_mesh();
  dxb = diff(c.xb);
  c.x0 = c.xb(1);
  c.unit_dx = min(dxb);
  nco = dxb/c.unit_dx;
  c.nc = round(nco);
  c.dipdeg = 12;
  c.mu = 1;
  c.nu = 0.25;
end

function xb = btrm_make_mesh ()
% Make a sample mesh. xb is element edges.
  xlim = [0 1];
  ffn = @(x) 1.25 + cos(2*pi*x);
  function f_min = fmfn (xlo, xhi)
    if (xlo <= 0.5 && xhi >= 0.5)
      f_min = ffn(0.5);
    else
      f_min = min(ffn(xlo), ffn(xhi));
    end
    f_min = 0.05*f_min;
  end
  xb = btrm_mesh(xlim, @fmfn);
end

function x = btrm_mesh (x, ffn)
% x = btrm_mesh(x, ffn)
%   ffn has the signature
%     f_min = ffn(xlo, xhi),
% where f_min = min_{x in [xlo xhi]}. f_min is smallest element size needed in
% the interval [xlo, xhi].
  o.max_height = inf;
  o.do_smooth = 1;
  o.max_dx = inf;
  assert(numel(x) == 2);
  assert(x(2) > x(1) && x(1) >= 0);
  cdr = @(x) x(2:end);
  % ffn is allowed to take just one argument x, element size at x, but then the
  % mesh is not guaranteed to be correct.
  good_mesher = nargin(ffn) == 2;
  function x = mesh1 (x, ffn)
  % Recursive mesher.
    if (good_mesher)
      split = ffn(x(1), x(2)) < diff(x);
    else
      x_sample = linspace(x(1), x(2), 10);
      split = any(ffn(x_sample) < diff(x));
    end
    if (split || diff(x) > o.max_dx)
      x_mid = 0.5*sum(x);
      x = [mesh1([x(1) x_mid], ffn) cdr(mesh1([x_mid, x(2)], ffn))];
    end
  end
  function [dx min_dx nc] = analyze (x);
  % Analyze the mesh cell sizes.
    dx = diff(x);
    min_dx = min(dx);
    nc = round(dx/min_dx);
  end
  % Make the initial mesh.
  x = mesh1(x, ffn);
  % If requested, limit the height of the mesh.
  if (~isempty(o.max_height))
    assert(o.max_height > 1);
    max_units = 2^(o.max_height - 1);
    [dx min_dx nc] = analyze(x);
    is = find(nc > max_units);
    pad = 0;
    for (i = is)
      k = pad + i;
      n = nc(i)/max_units;
      x = [x(1:k-1) linspace(x(k), x(k+1), n + 1) x(k+2:end)];
      pad = pad + n - 1;
    end
  end
  if (o.do_smooth)
    % Subdivide cells so that two adjacent cells differ by no more than one
    % level.
    [dx min_dx nc] = analyze(x);
    while (true)
      i = find(abs(diff(log2(nc))) > 1, 1);
      if (isempty(i)) break; end
      if (nc(i+1) > nc(i))
        nc = [nc(1:i) nc(i+1)/2*[1 1] nc(i+2:end)];
      else
        nc = [nc(1:i-1) nc(i)/2*[1 1] nc(i+1:end)];
      end
    end
    assert(all(round(nc) == nc));
    x = x(1) + min_dx*cumsum([0 nc]);
  end
end

function [Gs Gn] = btrm_call_hmmvp (c)
  gp = get_global_parms();
  dr = [gp.odir '/'];
  k.write_hmat_filename = [dr 'Gs'];
  k.tol = 1e-9;
  k.nthreads = 4;
  k.allow_overwrite = 1;
  k.greens_fn = 'btrmbem';
  k = transfer_fields(k, c, {'mu', 'nu', 'dipdeg', 'x0', 'unit_dx', 'nc'});
  k.component = 's';
  kvf_fn = [dr 'Gs.kvf'];
  kvf('Write', kvf_fn, k, 1);
  pr('--> hmmvp Gs\n');
  tic;
  system(sprintf('%s %s', gp.hmmvpbuild, kvf_fn));
  et = toc;
  pr('--> et %1.2f\n', et);
  Gs = btrm_read_hm(c, k);
  
  k.Bfro = norm(Gs, 'fro');
  k.write_hmat_filename = [dr 'Gn'];
  k.component = 'n';
  kvf_fn = [dr 'Gn.kvf'];
  kvf('Write', kvf_fn, k, 1);
  pr('--> hmmvp Gn\n');
  tic;
  system(sprintf('%s %s', gp.hmmvpbuild, kvf_fn));
  et = toc;
  pr('--> et %1.2f\n', et);
  Gn = btrm_read_hm(c, k);
end

function G = btrm_read_hm (c, k)
  [id nnz] = hmmvp('init', k.write_hmat_filename);
  G = hmmvp('extract', id, 1:numel(c.xb)-1, 1:numel(c.xb));
  hmmvp('reorganizememory', id); % just for testing
  pr('--> cf %1.2f\n', numel(c.xb)*(numel(c.xb) - 1)/nnz);
  hmmvp('cleanup', id);
end

% -- private
function t = tm_transform (t, dipdeg, min_depth)
% Transform the mesh so that it has minimum depth min_depth and dips at dipdeg.
  xy0 = min(t.v(1:2,:)');
  % Center in coord system.
  for (i = 1:2)
    t.v(i,:) = t.v(i,:) - xy0(i);
  end  
  % Rotate dip.
  cd = cosd(dipdeg); sd = sind(dipdeg);
  R = [cd -sd; sd cd];
  t.v(2:3,:) = R*t.v(2:3,:);
  t.v(3,:) = t.v(3,:) - max(t.v(3,:)) - min_depth;
  % Bring lower-left corner back to where it was.
  for (i = 1:2)
    t.v(i,:) = t.v(i,:) - min(t.v(i,:)) + xy0(i);
  end
  t = tm_calc_centroid(t);
end

function t = tm_calc_centroid (t)
  t.centroid = zeros(3, size(t.tri,2));
  for (i = 1:size(t.tri,2))
    t.centroid(:,i) = sum(t.v(:,t.tri(:,i)), 2)/3;
  end
end

function [fn cf] = tm_compress (c)
  % Compress with estimated ||B||_F.
  tmp_fn = make_tmp_filename();
  k = tm_make_kvf(c.o, c.t, tmp_fn);
  k.write_hd_filename = tmp_fn;
  run_hmmvp_build(k);
  % Get essentially exact ||B||_F.
  id = hmmvp('init', k.write_hmat_filename);
  Bfro = sqrt(hmmvp('fronorm2', id));
  hmmvp('cleanup', id);
  % Recompress.
  fn = tm_make_filename(c.o);
  k = tm_make_kvf(c.o, c.t, fn);
  k.use_hmat_filename = tmp_fn;
  k.use_hd_filename = tmp_fn;
  k.Bfro = Bfro;
  cf = run_hmmvp_build(k);
  % Remove temporary files.
  rm_tmp_files(tmp_fn);
end

function c = tm_make_kvf (o, t, fn)
  gp = get_global_parms();
  c = transfer_fields([], o, {'want_fullspace' 'disl' 'rcv'});
  c.tol = o.tol;
  c.nthreads = gp.nthreads;
  c.write_hmat_filename = fn;
  c.greens_fn = 'tgf';
  c.mu = 1; c.nu = 0.25;
  c.vertices = t.v;
  c.triangles = t.tri;
end

function fn = tm_make_filename (o)
  gp = get_global_parms();
  fn = sprintf(rmsp('%s/tm_ rfac%1.1f tol%d'), ...
               gp.odir, o.rfac, -round(log10(o.tol)));
end

% ------------------------------------------------------------------------------
% Utils.

function fn = make_tmp_filename ()
% Make a temporary filename.
  gp = get_global_parms();
  k = 0;
  %todo I could use random numbers so a linear search isn't needed.
  for (k = 0:1000)
    fn = [gp.odir filesep 'tmp' num2str(k)];
    d = dir([fn '.*']);
    if (isempty(d)) return; end
  end
  error('Failed to find a tmp filename.');
end

function rm_tmp_files (tmp_fns)
% Remove temporary files.
  if (~iscell(tmp_fns)) tmp_fns = {tmp_fns}; end
  exts = {'hd', 'hm', 'Bfroest', 'kvf'};
  for (i = 1:numel(tmp_fns))
    for (j = 1:numel(exts))
      fn = sprintf('%s.%s', tmp_fns{i}, exts{j});
      if (exist(fn, 'file') == 2)
        system(sprintf('rm -f %s;', fn));
      end
    end
  end
end

function [cf et] = run_hmmvp_build (k)
  gp = get_global_parms();
  kvf('Write', [k.write_hmat_filename '.kvf'], k, 1);

  cmd = sprintf('%s %s.kvf', gp.hmmvpbuild, k.write_hmat_filename);
  pr('hmex: %s\n', cmd);
  
  t = tic();
  system(cmd);
  et = toc(t);

  if (~isfield(k, 'write_Bfro_only') || ~k.write_Bfro_only)
    % Calculate the compression factor relative to a single-precision dense
    % matrix of this size. Include metadata, not just nnz.
    if (isfield(k, 'src_idxs'))
      mem_full = 4*numel(k.src_idxs)*numel(k.rcv_idxs);
    elseif (isfield(k, 'triangles'))
      mem_full = 4*size(k.triangles, 2)^2;
    else
      mem_full = 0;
    end
    d = dir([k.write_hmat_filename '.hm']);
    mem_hm = d.bytes;
    cf = mem_full/mem_hm;
  else
    cf = 0;
  end

  pr('hmex: wall-clock time %1.1f sec\n', et);
  if (cf) pr('hmex: compression factor %1.2f\n', cf); end
end

function slip = tapered_slip_fn (x, y, R, p)
  r2 = x.^2 + y.^2;
  slip = (R - r2).^p;
  slip(r2 > R) = 0;
end

function tsi = tsi_make (t, v)
  if (isstruct(t))
    tsi = TriScatteredInterp(t.centroid(1:2,:)', v(:));
  else
    tsi = TriScatteredInterp(t, v(:));
  end
  tsi.Method = 'natural'; 'linear';
end
function I = tsi_interp (tsi, x, y)
  [X Y] = meshgrid(x, y);
  xy = [X(:) Y(:)];
  I = reshape(tsi(xy), numel(y), numel(x));
end

% ------------------------------------------------------------------------------
% Standard utils.

function sd = transfer_fields (sd, ss, flds)
  if (~iscell(flds)) flds = {flds}; end
  for (i = 1:numel(flds))
    if (isfield(ss, flds{i})) sd.(flds{i}) = ss.(flds{i}); end
  end
end
function c = CC (v)
% Cell-centered from node-centered.
  c = 0.5*(v(1:end-1) + v(2:end));
end
function varargout = sp (varargin)
  [varargout{1:nargout}] = subplot(varargin{:});
end
function varargout = pr (varargin)
  fprintf(1, varargin{:});
end
function s = rmsp (s)
  s(s == ' ') = [];
end
function varargout = fi (varargin)
  if (nargin == 1)
    try
      % Don't let the figure window steal focus when the JVM is running.
      set(0, 'CurrentFigure', varargin{1});
    catch
      % Something went wrong (probably the figure doesn't exist), so go with
      % the default behavior.
      [varargout{1:nargout}] = figure(varargin{:});
    end
  else
    [varargout{1:nargout}] = figure(varargin{:});
  end
end
function cb (), colorbar; end

% ------------------------------------------------------------------------------
% Internal testing.

function oms = ok_t1 ()
  oms = {ok_mesh_rectangle(3, 2.5, 13, 5, [-2 1 -15], 20, 60), ...
         ok_mesh_rectangle(1.5, 2, 8, 4, [1 -1 -13], -10, 80)};
  ok_drawmesh(oms);
  xlabel('x'); ylabel('y'); zlabel('z');
  axis equal; grid on;
end

function B = image_hm (hm_fn)
  id = hmmvp('init', hm_fn);
  m = hmmvp('getm', id);
  n = hmmvp('getn', id);
  if (m*n <= 4000^2)
    B = hmmvp('extract', id, 1:m, 1:n);
    imagesc(log10(abs(B)));
  else
    B = [];
    pr('Matrix is too big.\n');
  end
  hmmvp('cleanup', id);
end

function c = ok_t2 ()
% Test GreensFnOkada for strike invariance in a halfspace.
  o = ok_setopts();
  o.vecs = o.vecs(:,:);
  ts = [0 -30];
  o.tol = 1e-30; % No compression.
  c = {};
  for (i = 1:numel(ts))
    origin = [0 0 -10]';
    oms = {ok_mesh_rectangle(2, 1, 20, 10, origin, -10 + ts(i), 20);
           ok_mesh_rectangle(3, 2, 18, 12, ...
                             ok_transform(origin, 0, ts(i), [5 -3 -5]'), ...
                             20 + ts(i), -30)};
    figure(1); ok_drawmesh(oms); hold all;
    c{i} = ok_build_hmatrices(oms, o);
    for (j = 1:numel(c{i}.hms))
      figure(2); c{i}.B{j} = image_hm(c{i}.hms{j});
      if (i == 2)
        pr('%2d %s re %10.2e norm %1.1e\n', ...
           j, c{1}.hms{j}, relerr(c{1}.B{j}, c{2}.B{j}), norm(c{1}.B{j}));
      end
    end
  end
end

function c = ok_t3 ()
% Test GreensFnOkada for strike and dip invariance in a fullspace. Because of
% the limitation on orientation of rectangles in dc3d, only a subset of
% transformations are possible.
  o = ok_setopts();
  o.vecs = o.vecs(:,:);
  % Has to be a fullspace because of changing dip and depth.
  o.want_fullspace = 1;
  o.tol = 1e-30; % No compression.
  ts = [0 -10];
  td = [0 20];
  c = {};
  origin = [0 0 -2]';
  for (i = 1:numel(ts))
    oms = {ok_mesh_rectangle(2, 1, 20, 10, origin, ts(i), 10 + td(i));
           ok_mesh_rectangle(3, 2, 18, 12, ...
                             ok_transform(origin, td(i), ts(i), [1 2 -3]'), ...
                             ts(i), -20 + td(i))};
    figure(1); ok_drawmesh(oms); hold all;
    c{i} = ok_build_hmatrices(oms, o);
    for (j = 1:numel(c{i}.hms))
      figure(2); c{i}.B{j} = image_hm(c{i}.hms{j});
      if (i == 2)
        co = c{1}.comps(j,:);
        msg = '';
        if (co(1) == co(2) && xor(co(3) == 3, co(4) == 3))
          % Freespace makes this component 0, so the relative error is
          % meaningless.
          msg = '(0 b/c fs)';
        end
        pr('%2d %s re %10.2e norm %1.1e %s\n', ...
           j, c{1}.hms{j}, relerr(c{1}.B{j}, c{2}.B{j}), norm(c{1}.B{j}), msg);
      end
    end
  end
end

function tm_t1 ()
  rfac = 1;
  xlim = [-1 1]; ylim = [-1 1];
  dx = diff(xlim)/(6*rfac); dy = diff(ylim)/(6*rfac);
  dipd = 12; min_depth = 2;
  types = [2 4];
  t = tm_mesh_rectangle(xlim, ylim, dx, dy, dipd, min_depth);
  tm_drawmesh(t); axis equal; axis tight;  
end
