/* hmmvpex: Examples for hmmvp
 *   Version 0.1
 *   Andrew M. Bradley
 *   ambrad@cs.stanford.edu
 *   CDFM Group, Geophysics, Stanford
 *   https://pangea.stanford.edu/research/CDFM/software
 * hmmvpex is licensed as follows:
 *   Open Source Initiative OSI - Eclipse Public License 1.0
 *   http://www.opensource.org/licenses/eclipse-1.0
*/

/* 'hmmvpbuild help compress' gives the following:

    btrmbem: 1D line fault in a 2D homogeneous elastic halfspace (plane
    strain). The mesh must be generated by recursively halving segments. I call
    this mesh a tree mesh.
      The geometry is as follows: x = 0 is at the surface. The mesh can start at
    x0 >= 0. If x0 = 0, the boundary condition is the free surface BC. If x0 >
    0, then the BC is that the fault has 0 slip (locked in the usual
    interpretation) in [0, x0]. The right side of the fault is at
        x_max = x0 + sum(nc)*unit_dx.
    The final column of G is for a constant-slip source extending from x_max to
    x = inf. This can be used to implement a velocity BC for x_max.
      The depth of the point x is  x sind(dipdeg).

      mu [scalar]: Shear modulus.
      nu [scalar]: Poisson's ratio.
      dipdeg [scalar]: Dip angle in degrees of the fault.
      x0 [scalar]: The updip point of the fault.
      unit_dx [scalar]: The size of the smallest element.
      nc [integer array]: Size of an element in terms of unit_dx. nc[i] must be
        a power of 2, and nc as a whole must describe a tree mesh.
      component ['s' or 'n']: Shear or normal.
*/

namespace btrmbem {
using namespace util;

namespace core {
// 'core' code is taken from btrmbem.c, which formed the whole matrix before
// using fdra's built-in, very old and suboptimal, H-matrix compressor.
// btrmbem.c was programmed in ANSI C89; I've made some changes in what follows.

#define cast(type, v) ((type)(v))
#define myassert(b) do {                                                \
    if (!(b)) fprintf(stderr, "--> False: " #b "\n"); } while (0)

inline void *mymalloc (size_t n)           { return malloc(n); }
inline void *myrealloc (void *p, size_t n) { return realloc(p, n); }
inline void  myfree (void *d)              { free(d); }
inline int imax (int a, int b) { return a >= b ? a : b; }
inline int imin (int a, int b) { return a <= b ? a : b; }

inline double GetSlope (const double x[2], const double y[2]) {
  return (y[1] - y[0]) / (x[1] - x[0]);
}

inline void GetQuadratic (
  double dx, /* Domain length. */
  double v1, bool have_s1, double s1, /* Value and possibly slope at left. */
  double v2, bool have_s2, double s2, /* At right. */
  double c[3]) /* Coeffs from x^2 to 1. */
{
  double dx2 = dx*dx;
  if (!have_s1) {
    double den = -dx2;
    double b1;
    c[2] = v1;
    b1 = v2 - c[2];
    c[0] = (b1 - dx*s2) / den;
    c[1] = (-2.0*dx*b1 + dx2*s2) / den;
  } else if (!have_s2) {
    c[2] = v1;
    c[1] = s1;
    c[0] = (v2 - c[1]*dx - c[2]) / dx2;
  } else {
    c[0] = c[1] = c[2] = 0.0;
    myassert(false);
  }
}

inline void GetCubic (double dx, double v1, double s1, double v2, double s2,
                      double c[4]) {
  double dx2 = dx*dx;
  double dx3 = dx2*dx;
  double den = -dx3;
  double b1, b2;
  c[2] = s1;
  c[3] = v1;
  b1 = v2 - dx*c[2] - c[3];
  b2 = s2 - c[2];
  c[0] = (2.0*b1 - dx*b2) / den;
  c[1] = (-3.0*dx*b1 + dx2*b2) / den;
}

static const UInt lci_stencil = 4;

/*todo Make a version that produces a sparse representation of non-0 yi. */
void LocalCubicInterp (const double *x, int nx, const double *y,
                       const double *xi, int nxi, double *yi) {
  const int nc = nx - 1;
  int i;
  int xi_i0, xi_i1;

  for (i = xi_i0 = xi_i1 = 0; i < nc; i++) {
    bool at_start = i == 0, at_end = i == nc - 1;

    /* Get the relevant part of xi: xi_i0:(xi_i1-1). */
    if (i > 0) xi_i0 = xi_i1;
    if (xi_i0 == nxi) break;
    if (at_end) xi_i1 = nxi;
    else {
      /*todo-opt Binary search. Even better, come up with a better start point
        than i = 0 before this loop. */
      while (xi[xi_i1] < x[i+1] && xi_i1 < nxi) xi_i1++;
    }

    /* If x == xi, set yi to y. */
    if (xi[xi_i0] == x[i]) {
      yi[xi_i0] = y[i];
      xi_i0++;
    }
    if (at_end && xi_i1 > xi_i0 + 1 && xi[xi_i1-1] == x[i+1]) {
      yi[xi_i1-1] = y[i+1];
      xi_i1--;
    }

    /* Interpolate the rest. */
    if (xi_i1 > xi_i0) {
      /*todo-maybe Could remember and shift using a 3-element array. */
      int j;
      double s1 = 0.0, s2 = 0.0;
      double smid = GetSlope(x+i, y+i);
      if (!at_start) {
        double a = (x[i] - x[i-1]) / (x[i+1] - x[i-1]);
        s1 = (1 - a)*GetSlope(x+i-1, y+i-1) + a*smid;
      }
      if (!at_end) {
        double a = (x[i+1] - x[i]) / (x[i+2] - x[i]);
        s2 = (1 - a)*smid + a*GetSlope(x+i+1, y+i+1);
      }

      if (at_start || at_end) {
        double c[3];
        GetQuadratic(
          x[i+1] - x[i], y[i], !at_start, s1, y[i+1], !at_end, s2, c);
        for (j = xi_i0; j < xi_i1; j++) {
          double xij = xi[j] - x[i];
          yi[j] = ((c[0]*xij + c[1])*xij) + c[2];
        }
      } else {
        double c[4];
        GetCubic(x[i+1] - x[i], y[i], s1, y[i+1], s2, c);
        for (j = xi_i0; j < xi_i1; j++) {
          double xij = xi[j] - x[i];
          yi[j] = (((c[0]*xij + c[1])*xij) + c[2])*xij + c[3];
        }
      }
    }
  }

  myassert(xi_i1 == nxi || yi[xi_i1] == y[nx-1]);
}

typedef struct {
  double mu, num1;
  double cd, c2d, sd, c4d, sd2, sd3;
} TractionMgr;

/* dip is in degrees. mu and nu are the Lame parameters. */
void trmgr_Init (TractionMgr *tm, double dip, double mu, double nu) {
  tm->mu = mu;
  tm->num1 = nu - 1.0;
  dip *= M_PI / 180.0;
  tm->cd  = cos(dip);
  tm->c2d = cos(2.0*dip);
  tm->c4d = cos(4.0*dip);
  tm->sd  = sin(dip);
  tm->sd2 = tm->sd * tm->sd;
  tm->sd3 = tm->sd2 * tm->sd;
}

static double Pow3 (double a) { return a*a*a; }

static void GetTraction2d_Formula (const TractionMgr *tm, double t, double z,
                                   double *Gs, double *Gn) {
  double t2 = t*t, t3 = t2*t, t4 = t3*t;
  double z2 = z*z, z3 = z2*z, z4 = z3*z;
  double den = M_PI * tm->num1 * Pow3(t2 + z2 - 2.0*t*z*tm->c2d);
  /*todo Horner form? */
  if (Gs)
    *Gs = (-2.0 * z * (t + z) * tm->mu * 
           (t4 - 3.0 * t3 * z + 7.0 * t2 * z2 - 3.0 * t * z3 + z4 +
            (t4 - 5.0 * t3 * z + 4.0 * t2 * z2 - 5.0 * t * z3 + z4) * tm->c2d +
            t2 * z2 * tm->c4d) * tm->sd2) /
      ((t - z) * den);
  if (Gn)
    *Gn = (4.0 * z2 * tm->mu * tm->cd *
           (-3.0 * t2 * z + z3 + 2.0 * t3 * tm->c2d) * tm->sd3) /
      den;
}

/* Get shear and normal tractions from line sources.
     xb is an array of line-segment source end points.
     xo is the receiver point.
     If want_vbc, then the velocity boundary condition is at xb(end) and is
   calculated. This requires Gs/n to have one more element available than there
   are cells in xb.
     Gs and Gn contain the shear and normal tractions. Results are *added* to Gs
   and Gn, so don't forget to zero them if necessary.
     stride is for Gs/n. Gs[i] is set for i = 0, stride, 2*stride,...  */
void GetTraction2d (const TractionMgr *tm, const double *xb, UInt nxb,
                    double xo, bool want_vbc, double *Gs, double *Gn,
                    UInt stride) {
  UInt i;
  double Gs1, Gs2, Gn1, Gn2;
  GetTraction2d_Formula(tm, xo, xb[0], &Gs1, &Gn1);
  for (i = 0; i < nxb - 1; i++) {
    GetTraction2d_Formula(tm, xo, xb[i+1], &Gs2, &Gn2);
    if (Gs) { *Gs += Gs2 - Gs1; Gs += stride; Gs1 = Gs2; }
    if (Gn) { *Gn += Gn2 - Gn1; Gn += stride; Gn1 = Gn2; }
  }
  if (want_vbc) {
    if (Gs) *Gs -= Gs1;
    if (Gn) *Gn -= Gn1;
  }
}

/* -----------------------------------------------------------------------------
 * Mesh.  */

typedef struct {
  const UInt *nc; /* number of unit cells in each cell */
  UInt nc_tot;    /* total number of unit cells in domain */
  UInt n;         /* number of cells */
  double *xb;     /* cell boundaries */
  double *x;      /* cell centers plus xb[0] and xb[end] for BCs */
} Domain;

void Domain_Init (Domain *d, const UInt *nc, UInt n, double x0,
                  double unit_dx) {
  UInt i;
  d->nc = nc;
  d->n = n;
  d->xb = (double*) mymalloc((n + 1)*sizeof(double));
  d->x = (double*) mymalloc((n + 2)*sizeof(double));
  d->xb[0] = x0;
  d->x[0] = x0;
  d->x++; /* [1] Make d->x[0] the first element center. */
  for (i = 0, d->nc_tot = 0; i < n; i++) {
    d->nc_tot += nc[i];
    d->xb[i+1] = x0 + d->nc_tot * unit_dx;
    d->x[i] = 0.5*(d->xb[i] + d->xb[i+1]);
  }
  d->x[n] = d->xb[n];
}

void Domain_Clear (Domain *d) {
  if (d->xb) myfree(d->xb);
  if (d->x) myfree(d->x - 1); /* -1 because of [1]. */
  d->x = d->xb = NULL;
  d->n = 0;
  d->nc = NULL;
}

typedef struct {
  UInt i;      /* cell id */
  UInt li, ri; /* cells li:ri are in the nbrhd */
  UInt ili;    /* cells used for interp only are ili:li-1 and ... */
  UInt iri;    /* ... ri:iri */
  UInt unit;   /* factor of unit cell that sets the unit size for this nbrhd */
} RcvrNbrhd;

typedef enum { rsr_Normal = 0, rsr_InNbrhd, rsr_InterpOnly } RcvrSrcRelation;

static void
nbr_SetMinNhszs (const Domain *dmn, double dist_fac,
                 UInt *nhszs /* numel == n */) {
  for (UInt i = 0; i < dmn->n; i++) nhszs[i] = 0;
  for (UInt i = 0; i < dmn->n; i++) {
    int j;
    UInt d;
    UInt dfnci = (UInt) ceil(dist_fac * dmn->nc[i]);

    if (dfnci > nhszs[i]) nhszs[i] = dfnci;

    d = 0;
    for (j = i+1; j < (int) dmn->n; j++) {
      d += dmn->nc[j];
      if (d > dfnci) break;
      if (dfnci > nhszs[j]) nhszs[j] = dfnci;
    }

    d = 0;
    for (j = i-1; j >= 0; j--) {
      d += dmn->nc[j];
      if (d > dfnci) break;
      if (dfnci > nhszs[j]) nhszs[j] = dfnci;
    }
  }
}

static void
nbr_GetNbrhd (RcvrNbrhd *rn, const Domain *dmn, UInt stencil, UInt nhsz) {
  UInt ld = 0, rd = 0;
  rn->li = rn->ri = rn->i;
  ld = rd = 0;
  rn->unit = dmn->nc[rn->i];
  for (;;) {
    bool do_lt, do_rt;
    { bool ls = ld >= nhsz || rn->li == 0;            /* Left far enough? */
      bool rs = rd >= nhsz || rn->ri == dmn->n;      /* Right far enough? */
      if (ls && rs) {
        /* Experimental. Seems to be working. */
        /*printf("[imb %d %d %d] ", rn->i, ld, rd);*/
        break;
      }
      do_lt = !ls || rs;
      do_rt = !rs || ls; }
    if (do_lt && do_rt) {
      /* If neither is satisfied, let only the smaller region grow. */
      if (ld < rd)      do_rt = false;
      else if (ld > rd) do_lt = false;
    }
    if (do_lt) {
      rn->li--;
      ld += dmn->nc[rn->li];
      if (dmn->nc[rn->li] < rn->unit) rn->unit = dmn->nc[rn->li];
    }
    if (do_rt) {
      rn->ri++;
      if (rn->ri < dmn->n) { /* Not the v-BC. */
        rd += dmn->nc[rn->ri];
        if (dmn->nc[rn->ri] < rn->unit) rn->unit = dmn->nc[rn->ri];
      }
    }
  }
  { UInt delta = stencil / 2;
    rn->ili = imax(0, (int) rn->li - delta);
    /* The right side incorporates the v BC, so we use dmn->n rather than
       dmn->n-1. */
    /*todo-bc */
    rn->iri = imin(dmn->n, rn->ri + delta);
  }
}

static void
nbr_GetNbrhds (const Domain *dmn, double dist_fac,
               RcvrNbrhd *rns /* numel(rns) == dmn->n */) {
  UInt i;
  UInt *nhszs;

  nhszs = cast(UInt*, mymalloc(dmn->n*sizeof(UInt)));
  nbr_SetMinNhszs(dmn, dist_fac, nhszs);

  for (i = 0; i < dmn->n; i++) {
    rns[i].i = i;
    nbr_GetNbrhd(rns+i, dmn, lci_stencil, nhszs[i]);
    if (dmn->nc[i] > rns[i].unit) {
      /* If nc[i] is larger than the others in its initial nbrhd, increase the
         size of the nbrhd, as I've found this case needs a better approximation
         than the opposite case. */
      nbr_GetNbrhd(rns+i, dmn, lci_stencil, 8 * nhszs[i]);
    }
  }

  myfree(nhszs);

#if 0
  if (dmn->n < 20) {
    for (i = 0; i < dmn->n; i++)
      printf("%d (%d, %d): %d %d %d %d\n", i, rns[i].unit, dmn->nc[rns[i].i],
             rns[i].ili, rns[i].li, rns[i].ri, rns[i].iri);
  }
#endif
}

static RcvrSrcRelation
nbr_GetRelation (const RcvrNbrhd *rn, UInt src_id) {
  if (src_id < rn->ili || src_id > rn->iri)
    return rsr_Normal;
  if (src_id >= rn->li && src_id <= rn->ri)
    return rsr_InNbrhd;
  return rsr_InterpOnly;
}

typedef struct {
  int interp;
  double dist_fac;
} BrickMeshBemBuilderOptions;

void opts_SetDefaults (BrickMeshBemBuilderOptions *o) {
  o->interp = 3;
  o->dist_fac = 4;
}

struct BrickMeshBemBuilder {
  const TractionMgr *tm;
  const Domain *d;
  const RcvrNbrhd *rns;
  /* Count the number of Gs calculations. In reality, for these 2D calculations,
     calculating one element and then another separately involves redundant
     calculations. But these are not our interest because dc3d does not work
     like that (at least without modification). */
  UInt ctr;
  double *y;   /* y is defined on CC(d->xb); CC = cell centered */
  double *xbi; /* xi = CC(xbi), yi are the interpolation */
  double *xi;
  double *yi;
  bool want_Gs, want_Gn;
  double *Gs;  /* work arrays */
  double *Gn;
  UInt nbi;    /* numbel(xbi) */
  UInt ni;     /* numel(xi) == numel(yi) == numel(Gs,n) */
  UInt nicap;  /* capacity of the array */
};

static void
bmb_Init (BrickMeshBemBuilder *b, const TractionMgr *tm, const Domain *d,
          const RcvrNbrhd* rns, const BrickMeshBemBuilderOptions* o,
          bool want_Gs, bool want_Gn) {
  b->tm = tm;
  b->d = d;
  b->rns = rns;

  b->ctr = 0;

  /* Make room for handling a 0v-BC on the updip side. If it's used, b->y[-1]
     is always 0 because it's the 0 slip speed BC. */
  b->y = (double*) mymalloc((d->n + 2)*sizeof(double));
  b->y[0] = 0;
  b->y++;

  b->xi = b->xbi = b->yi = b->Gs = b->Gn = NULL;
  b->nbi = b->ni = b->nicap = 0;
  b->want_Gs = want_Gs;
  b->want_Gn = want_Gn;
}

static void bmb_Clear(BrickMeshBemBuilder *b) {
  b->tm = NULL;
  b->d = NULL;
  b->rns = NULL;
  b->ctr = 0;
  if (b->y) myfree(b->y - 1);
  if (b->xi) myfree(b->xi);
  if (b->xbi) myfree(b->xbi);
  if (b->yi) myfree(b->yi);
  if (b->Gs) myfree(b->Gs);
  if (b->Gn) myfree(b->Gn);
  b->y = b->xi = b->xbi = b->yi = b->Gs = b->Gn = NULL;
  b->nbi = b->ni = b->nicap = 0;
}

static void bmb_AllocSlot (BrickMeshBemBuilder *b) {
  if (b->ni + 1 > b->nicap) {
    b->nicap = b->nicap ? 2*b->nicap : 1;
    b->xbi = (double*) myrealloc(b->xbi, (b->nicap + 1)*sizeof(double));
    b->xi = (double*) myrealloc(b->xi, b->nicap*sizeof(double));
    b->yi = (double*) myrealloc(b->yi, b->nicap*sizeof(double));
    if (b->want_Gs)
      b->Gs = (double*) myrealloc(b->Gs, b->nicap*sizeof(double));
    if (b->want_Gn)
      b->Gn = (double*) myrealloc(b->Gn, b->nicap*sizeof(double));
  }
}

static void bmb_InitX (BrickMeshBemBuilder *b, double xb0) {
  b->ni = 0;
  bmb_AllocSlot(b);
  b->xbi[0] = xb0;
  b->nbi = 1;
}

static void bmb_AppendX (BrickMeshBemBuilder *b, double xbi) {
  bmb_AllocSlot(b);
  b->xi[b->ni] = 0.5*(b->xbi[b->ni] + xbi);
  b->ni++;
  b->xbi[b->nbi] = xbi;
  b->nbi++;
}

/*todo-bc */
static void bmb_AppendBcX (BrickMeshBemBuilder *b) {
  bmb_AllocSlot(b);
  b->xi[b->ni] = b->xbi[b->nbi-1]; /* comes from last call to bmb_AppendX */
  b->ni++;
}

static void bmb_RmvLeadingTrailing0s (BrickMeshBemBuilder *b) {
  int i, n0_fst = (int) b->ni, n0_lst = -1;
  bool fnd = false;
  /* Shrink to the non-0 region. */
  for (i = 0; i < (int) b->ni; i++)
    if (b->yi[i] != 0.0) { n0_fst = i; fnd = true; break; }
  for (i = (int) b->ni - 1; i >= 0; i--)
    if (b->yi[i] != 0.0) { n0_lst = i; fnd = true; break; }
  myassert(n0_fst < (int) b->ni && n0_lst >= 0);
  if (!fnd) return;
  /* Clean up xi, yi. */
  if (n0_fst > 0)
    for (i = n0_fst; i <= n0_lst; i++) {
      /* We don't really have to make xi consistent, as it's not used
         subsequently. */
      b->xi[i - n0_fst] = b->xi[i];
      b->yi[i - n0_fst] = b->yi[i];
    } 
  b->ni = n0_lst - n0_fst + 1;
  /* Now handle xbi. */
  for (i = 0; i < (int) b->nbi; i++)
    if (b->xbi[i] > b->xi[0]) { n0_fst = i - 1; break; }
  myassert(n0_fst >= 0 && n0_fst <= (int) b->nbi - 1 &&
           b->xbi[n0_fst] <= b->xi[0]);
  for (i = (int) b->nbi - 1; i >= 0; i--)
    if (b->xbi[i] < b->xi[b->ni-1]) { n0_lst = i + 1; break; }
  myassert(n0_lst > 0 && n0_lst <= (int) b->nbi - 1 &&
           b->xbi[n0_lst] >= b->xi[b->ni - 1]);
  if (n0_fst > 0)
    for (i = n0_fst; i <= n0_lst; i++)
      b->xbi[i - n0_fst] = b->xbi[i];
  b->nbi = n0_lst - n0_fst + 1;
}

static double bmb_Dot (const BrickMeshBemBuilder *b, const double *v) {
  UInt i;
  double d = 0;
  for (i = 0; i < b->ni; i++) d += b->yi[i] * v[i];
  return d;
}

static void
bmb_GetXbiYi (BrickMeshBemBuilder *b, const RcvrNbrhd *rn, UInt src_id) {
  UInt i;
  const Domain *d = b->d;
  /* Determine the region of x that src_id supports. */
  /*todo-bc */
  const UInt delta = lci_stencil/2;
  /*todo (1) If we decide to make Interp take all_0_lt, all_0_rt arguments, then
    we can tighten this region. For now, we need Interp not to switch to its
    boundary method. */
  const UInt x_ifst = imax(0, (int) src_id - delta - 1);
  const UInt x_ilst = imin(d->n, src_id + delta + 1);
  double *x = d->x + x_ifst;
  UInt nx = x_ilst - x_ifst + 1;
  double* y = b->y;

  /* Determine xbi, xi in this region. */
  bmb_InitX(b, d->xb[x_ifst]);
  for (i = x_ifst; i <= x_ilst; i++) {
    if (i == d->n) {
      /*todo-bc */
      bmb_AppendBcX(b);
    } else {
      RcvrSrcRelation rsr = nbr_GetRelation(rn, i);
      if (rsr == rsr_InterpOnly ||
          rsr == rsr_Normal || /* b/c of (1) */
          d->nc[i] == rn->unit) {
        bmb_AppendX(b, d->xb[i+1]);
      } else {
        UInt j;
        const UInt nsub = d->nc[i] / rn->unit;
        const double dx = (d->xb[i+1] - d->xb[i]) / (double) nsub;
        myassert(nsub * rn->unit == d->nc[i]);
        for (j = 1; j <= nsub; j++)
          bmb_AppendX(b, d->xb[i] + j*dx);
      }
    }
  }

  /* Make y = e_i in this region. */
  memset(y, 0, nx*sizeof(double));
  y[src_id - x_ifst] = 1;

  /* Interp y. */
  if (x_ifst == 0 && d->xb[0] > 0) {
    /* Make the 0-v BC at the updip end available. This prevents the
       interpolator from switching to its extrapolation method. */
    x--;
    y--;
    nx++;
  }
  LocalCubicInterp(x, nx, y, b->xi, b->ni, b->yi);

  /* Tighten xi, xbi, yi so that yi has no leading or trailing 0s. */
  bmb_RmvLeadingTrailing0s(b);
}

static void
bmb_GetTraction2d (BrickMeshBemBuilder *b, UInt rcv_id, UInt src_id,
                   double *Gs, double *Gn) {
  const RcvrNbrhd *rn = b->rns + rcv_id;
  const Domain *d = b->d;
  bool is_boundary = src_id == d->n; /*todo-bc */
  if (Gs) *Gs = 0;
  if (Gn) *Gn = 0;
  RcvrSrcRelation rsr = nbr_GetRelation(rn, src_id);
  if (rsr == rsr_Normal) {
    GetTraction2d(b->tm, d->xb + src_id, is_boundary ? 1 : 2,
                  0.5*(d->xb[rn->i] + d->xb[rn->i+1]), is_boundary,
                  Gs, Gn, 1);
    b->ctr += 1 + (is_boundary ? 1 : 0);
  } else {
    UInt ir;
    const UInt nsub = d->nc[rn->i] / rn->unit;
    const double scale = 1.0 / (double) nsub;
    const double dxh = 0.5 * scale * (d->xb[rn->i+1] - d->xb[rn->i]);
    myassert(rn->unit * nsub == d->nc[rn->i]);
    bmb_GetXbiYi(b, rn, src_id);
    if (b->Gs) memset(b->Gs, 0, b->ni*sizeof(double));
    if (b->Gn) memset(b->Gn, 0, b->ni*sizeof(double));
    for (ir = 0; ir < nsub; ir++) {
      double xo = d->xb[rn->i] + (1 + 2*ir) * dxh;
      GetTraction2d(b->tm, b->xbi, b->nbi, xo, is_boundary, b->Gs, b->Gn, 1);
    }
    if (Gs) *Gs += scale * bmb_Dot(b, b->Gs);
    if (Gn) *Gn += scale * bmb_Dot(b, b->Gn);
    b->ctr += nsub * (b->nbi - 1 + (is_boundary ? 1 : 0));
  }
}

#undef cast
#undef myassert
} // core

typedef Matrix<double> Matd;

class GreensFn : public ImplGreensFn {
public:
  virtual ~GreensFn();
  virtual void Init(const KeyValueFile* kvf) throw (Exception);
  virtual Hd* ComputeHd(double eta);
  virtual bool Call(const CompressBlockInfo& cbi, const vector<UInt>& rs,
                    const vector<UInt>& cs, double* B) const;

private:
  struct Component { enum Enum { shear, normal }; };
  Component::Enum component;

  struct {
    vector<UInt> nc;
    vector<core::RcvrNbrhd> rns;
    core::TractionMgr tm;
    core::Domain d;
    mutable vector<core::BrickMeshBemBuilder> b;
  } bmb;
};

Hd* GreensFn::ComputeHd (double eta) {
  // Compute element centers in the obvious rotated and scaled space. (The
  // spatial decomposition in exact arithmetic is invariant to these
  // transformations.)
  Matd xc(3, bmb.nc.size());
  xc.Bzero();
  UInt x = 0;
  for (size_t i = 0; i < bmb.nc.size(); i++) {
    xc(1, i+1) = x + 0.5*bmb.nc[i];
    x += bmb.nc[i];
  }

  // Add an extra point far away for the v BC.
  Matd src = xc;
  src.Reshape(3, bmb.nc.size() + 1);
  src(1, bmb.nc.size() + 1) = 10*src(1, bmb.nc.size());

  return NewHd(src, xc, NULL, eta);
}

bool GreensFn::Call (const CompressBlockInfo& cbi, const vector<UInt>& rs,
                     const vector<UInt>& cs, double* B) const {
  size_t tid = omp_get_thread_num();
  for (UInt k = 0, ic = 0; ic < cs.size(); ic++) {
    const UInt c = cs[ic] - 1;
    for (UInt ir = 0; ir < rs.size(); ir++, k++) {
      const UInt r = rs[ir] - 1;
      double *d1 = NULL, *d2 = NULL;
      if (component == Component::shear) d1 = B + k; else d2 = B + k;
      core::bmb_GetTraction2d(&bmb.b[tid], r, c, d1, d2);
    }
  }
  return true;
}

// Read data from the key-value files.
#define readmat(name, var) do {                                         \
    if (!kvf->GetMatd(#name, m)) throw Exception("Missing " #name);     \
    var = *m;                                                           \
  } while (0)
#define readstr(name, var) do {                                         \
    if (!kvf->GetString(#name, s)) throw Exception("Missing " #var);    \
    var = *s;                                                           \
  } while (0)
#define readchar(var) do {                                              \
    if (!kvf->GetString(#var, s)) throw Exception("Missing " #var);     \
    if (s->size() != 1) throw Exception("Must be a char: " #var);       \
    var = (*s)[0];                                                      \
  } while (0)
#define readdbl(var) do {                                               \
    if (!kvf->GetDouble(#var, d)) throw Exception("Missing " #var);     \
    var = d;                                                            \
  } while (0)
#define checkvarsz2(var, m, n) do {             \
    if (var.Size(1) != m || var.Size(2) != n)   \
      throw Exception("Wrong size: " #var);     \
  } while (0)
#define checkvarsz1(var, m) do {                \
    if (var.Size() != m)                        \
      throw Exception("Wrong size: " #var);     \
  } while (0)

inline bool IsPow2 (UInt n) { return !(n & (n - 1)); }

// Read the key-value file and initialize data.
void GreensFn::Init (const KeyValueFile* kvf) throw (Exception) {
  const Matd* m;
  const string* s;
  double d;

  double mu, nu, dipdeg, x0, unit_dx;
  readdbl(mu);
  readdbl(nu);
  readdbl(dipdeg);
  readdbl(x0); if (x0 < 0) throw Exception("x0 must be >= 0.");
  readdbl(unit_dx); if (unit_dx <= 0) throw Exception("unit_dx must be > 0.");

  size_t nthreads = 1;
  if (mpi::GetNproc() == 1 && kvf->GetDouble("nthreads", d) && d >= 1)
    nthreads = (size_t) d;

  component = Component::shear;
  { string c;
    readstr(component, c);
    if (c == "n")
      component = Component::normal;
    else if (c == "s")
      component = Component::shear;
    else
      throw Exception("component must be 's' or 'n'."); }

  { Matd mnc;
    readmat(nc, mnc);
    bmb.nc.resize(mnc.Size());
    for (size_t i = 0; i < bmb.nc.size(); i++) {
      bmb.nc[i] = (UInt) mnc(i+1);
      if (bmb.nc[i] == 0 || !IsPow2(bmb.nc[i]))
        throw Exception("nc[i] must be 2^k for k >= 0.");
    }}

  core::trmgr_Init(&bmb.tm, dipdeg, mu, nu);
  core::Domain_Init(&bmb.d, &bmb.nc[0], bmb.nc.size(), x0, unit_dx);
  core::BrickMeshBemBuilderOptions o;
  core::opts_SetDefaults(&o);
  // Always use inf, as generally we can afford it on 1D faults.
  o.dist_fac = bmb.d.nc_tot + 1;

  bmb.rns.resize(bmb.d.n);
  core::nbr_GetNbrhds(&bmb.d, o.dist_fac, &bmb.rns[0]);
  bmb.b.resize(nthreads);
  for (size_t i = 0; i < nthreads; i++)
    core::bmb_Init(&bmb.b[i], &bmb.tm, &bmb.d, &bmb.rns[0], &o,
                   component == Component::shear,
                   component == Component::normal);
}

#undef checkvarsz1
#undef checkvarsz2
#undef readdbl
#undef readstr
#undef readmat

GreensFn::~GreensFn () {
  for (size_t i = 0; i < bmb.b.size(); i++) core::bmb_Clear(&bmb.b[i]);
  core::Domain_Clear(&bmb.d);
}
}
