/* hmmvpex: Examples for hmmvp
 *   Version 0.1
 *   Andrew M. Bradley
 *   ambrad@cs.stanford.edu
 *   CDFM Group, Geophysics, Stanford
 *   https://pangea.stanford.edu/research/CDFM/software
 * hmmvpex is licensed as follows:
 *   Open Source Initiative OSI - Eclipse Public License 1.0
 *   http://www.opensource.org/licenses/eclipse-1.0
*/

/* Example Green's function implementation that wraps Okada's routine DC3D:
       Y. Okada, Internal deformation due to shear and tensile faults in a
       half-space, Bull. Seism. Soc. Am., 1992.

   Disclaimer: This file is meant to be an example of how to insert new
   ImplGreensFns into hmmvp. It is not very well tested.

   'hmmvpbuild help compress' gives the following:

    okada: Wrapper to DC3D [Okada 1992], which is used to compute traction due
           to constant displacement on a rectangle in a homogeneous elastic
           halfspace.
      mu [scalar]: Shear modulus.
      nu [scalar]: Poisson's ratio.
      center [3xNr array]: Rectangle centers in global coordinates. Nr is the
        number of rectangles. -Z is below the surface, so all Z values should
        be <= 0.
      striked [Nr array]: Strike in degrees of a rectangle.
      dipd [Nr array]: Dip in degrees of a rectangle. At 0 strike, the dip
        direction is Y, the strike direction is X, and the element normal is
        +Z. Positive dipd moves the +Y side in the +Z direction. Positive
        striked rotates clockwise about the +Z vector.
      len_along_dip [Nr array]: Length of a rectangle in the dip direction.
      len_along_strike [Nr array]: Length of a rectangle in the strike
        direction.
      disl [3xNr array]: Dislocation vectors in local coordinates. x component
        is strike-slip, y is dip-slip, z is tensile.
      rcv [3xNr array]: Traction vectors in local coordinates.
      src_idxs [array]: Index (starting at 1) set listing source rectangles.
      rcv_idxs [array]: Index set listing receiver rectangles. src_idxs and
        rcv_idxs can overlap. The matrix G is ordered according to src_idxs and
        rcv_idxs: the columns are the sources in the order of src_idxs, and the
        rows are the receivers in the order of rcv_idxs.
      want_fullspace [scalar, optional, default 0]: Non-0 if a fullspace, rather
        than a halfspace, is desired.
 */

#include "Elastostatics.hpp"

namespace okada {
using namespace util;
using namespace es;
using namespace dc3;

typedef Matrix<double> Matd;

class GreensFn : public ImplGreensFn {
public:
  virtual void Init(const KeyValueFile* kvf) throw (Exception);
  virtual Hd* ComputeHd(double eta);
  virtual bool Call(const CompressBlockInfo& cbi, const vector<UInt>& rs,
                    const vector<UInt>& cs, double* B) const;

private:
  vector<UInt> src_idxs, rcv_idxs;
  Matd ctr, normal, src_lcl, rcv_lcl, al, aw, striked, dipd;
  LameParms lp;
  bool want_fullspace;
};

// y += a x.
inline void axpy3 (const double a, const double x[3], double y[3]) {
  y[0] += a*x[0]; y[1] += a*x[1]; y[2] += a*x[2];
}
// dot(a, b)
inline double dot3 (const double a[3], const double b[3]) {
  return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}

// v = [c s; -s c] v, where c = cos(strike), s = sin(strike).
inline void
rotate_striked (const double striked, double* v, const size_t nv) {
  const double c = cosd(striked), s = sind(striked);
  for (size_t i = 0; i < nv; i++) {
    const double t[2] = {v[0], v[1]};
    v[0] =  c*t[0] + s*t[1];
    v[1] = -s*t[0] + c*t[1];
    v += 3;
  }
}

void copy_ctr (Matd& cctr, const Matd& ctr, const vector<size_t>& idxs) {
  cctr.Resize(3, idxs.size());
  for (size_t i = 0; i < idxs.size(); i++)
    for (size_t j = 1; j <= 3; j++)
      cctr(j, i+1) = ctr(j, idxs[i]);
}

// Compute the spatial decomposition of the point cloud of rectangle centers.
Hd* GreensFn::ComputeHd (double eta) {
  Matrix<double> rctr(3, rcv_idxs.size()), sctr(3, src_idxs.size());
  copy_ctr(rctr, ctr, rcv_idxs);
  copy_ctr(sctr, ctr, src_idxs);
  return NewHd(sctr, rctr, NULL, eta);
}

inline void calc_rcv_vectors (
  const double dipd, const double rel_striked, const double rcv_lcl[3],
  double normal[3], double rcv_glb[3])
{
  const double cd = cosd(dipd), sd = sind(dipd);
  double y[9] = {1, 0, 0,  0, cd, sd,  0, -sd, cd};
  rotate_striked(rel_striked, y, 3);
  memcpy(normal, y+6, 3*sizeof(double));
  memset(rcv_glb, 0, 3*sizeof(double));
  axpy3(rcv_lcl[0], y,   rcv_glb);
  axpy3(rcv_lcl[1], y+3, rcv_glb);
  axpy3(rcv_lcl[2], y+6, rcv_glb);
}

// Calculate normal' Sigma v, where Sigma is the stress tensor.
inline double
calc_traction (const double normal[3], const double v[3], const double s[6]) {
  double ns[3];
  ns[0] = normal[0]*s[0] + normal[1]*s[1] + normal[2]*s[2];
  ns[1] = normal[0]*s[1] + normal[1]*s[3] + normal[2]*s[4];
  ns[2] = normal[0]*s[2] + normal[1]*s[4] + normal[2]*s[5];
  return dot3(ns, v);
}

// Calculate B = G(rs, cs), where G is the array of Green's functions.
bool GreensFn::Call (const CompressBlockInfo& cbi, const vector<UInt>& rs,
                     const vector<UInt>& cs, double* B) const {
  for (UInt k = 0, ic = 0; ic < cs.size(); ic++) {
    const UInt c = src_idxs[cs[ic] - 1];
    for (UInt ir = 0; ir < rs.size(); ir++, k++) {
      const UInt r = rcv_idxs[rs[ir] - 1];
      double stress[6];
      { double u[3], du[9];
        // Receiver element center ...
        double xo[3] = { ctr(1,r) - ctr(1,c), ctr(2,r) - ctr(2,c), ctr(3,r) };
        // ... rotated into dc3d source element coordinates.
        rotate_striked(-striked(c), xo, 1);
        // Call Okada's DC3D.
        Dc3d(lp.alpha(),
             xo[0], xo[1], xo[2],
             -ctr(3,c), dipd(c),
             al(c), al(c), aw(c), aw(c),
             src_lcl(1,c), src_lcl(2,c), src_lcl(3,c),
             u, du,
             want_fullspace);
        // Calculate the stress tensor given displacement gradient in source
        // element coordinates.
        DuToS(lp, du, stress); }
      // Calculate the receiver vectors in source element coordinates.
      double normal[3], rcv_src_elem[3];
      calc_rcv_vectors(dipd(r), striked(r) - striked(c), &rcv_lcl(1,r),
                       normal, rcv_src_elem);
      // Calculate traction on the receiver.
      B[k] = calc_traction(normal, rcv_src_elem, stress);
    }
  }
  return true;
}

// Read data from the key-value files.
#define readmat(name, var) do {                                         \
    if (!kvf->GetMatd(#name, m)) throw Exception("Missing " #name);     \
    var = *m;                                                           \
  } while (0)
#define readstr(var) do {                                               \
    if (!kvf->GetString(#var, s)) throw Exception("Missing " #var);     \
    var = *s;                                                           \
  } while (0)
#define readchar(var) do {                                              \
    if (!kvf->GetString(#var, s)) throw Exception("Missing " #var);     \
    if (s->size() != 1) throw Exception("Must be a char: " #var);       \
    var = (*s)[0];                                                      \
  } while (0)
#define readdbl(var) do {                                               \
    if (!kvf->GetDouble(#var, d)) throw Exception("Missing " #var);     \
    var = d;                                                            \
  } while (0)
#define checkvarsz2(var, m, n) do {             \
    if (var.Size(1) != m || var.Size(2) != n)   \
      throw Exception("Wrong size: " #var);     \
  } while (0)
#define checkvarsz1(var, m) do {                \
    if (var.Size() != m)                        \
      throw Exception("Wrong size: " #var);     \
  } while (0)

void copy_idxs (vector<UInt>& idxs, const Matd& aidxs, const UInt imax)
  throw (Exception)
{
  idxs.resize(aidxs.Size());
  for (size_t i = 0; i < idxs.size(); i++) {
    idxs[i] = (UInt) aidxs(i+1);
    if (idxs[i] > imax || idxs[i] < 1)
      throw Exception("Index is out of bound.");
  }
}

// Read the key-value file and initialize data.
void GreensFn::Init (const KeyValueFile* kvf) throw (Exception) {
  const Matd* m;
  double d;

  double mu, nu;
  readdbl(mu);
  readdbl(nu);
  lp.Set(mu, nu);

  readmat(center, ctr);
  const size_t nr = ctr.Size(2);
  checkvarsz2(ctr, 3, nr);

  readmat(len_along_dip, aw); checkvarsz1(aw, nr); aw *= 0.5;
  readmat(len_along_strike, al); checkvarsz1(al, nr); al *= 0.5;
  readmat(dipd, dipd); checkvarsz1(dipd, nr);
  readmat(striked, striked); checkvarsz1(striked, nr);
  readmat(disl, src_lcl); checkvarsz2(src_lcl, 3, nr);
  readmat(rcv, rcv_lcl); checkvarsz2(rcv_lcl, 3, nr);

  { Matrix<double> idxs;
    readmat(src_idxs, idxs);
    copy_idxs(src_idxs, idxs, nr);
    readmat(rcv_idxs, idxs);
    copy_idxs(rcv_idxs, idxs, nr); }

  want_fullspace = false;
  if (kvf->GetDouble("want_fullspace", d)) want_fullspace = (bool) d;
}

#undef checkvarsz1
#undef checkvarsz2
#undef readdbl
#undef readstr
#undef readmat

#undef bapr
}
